﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.IO.Ports;
using System.Drawing.Imaging;

using Datecs.Api.IO;
using Datecs.Api.Card;
using Datecs.Api.Printer;
using Datecs.Api.UniversalReader;
using Datecs.Api.SmartCard;
using Datecs.Api.Util;

namespace DatecsPrint
{
    public partial class MainForm : Form
    {
        // Helper class that implements file logger
        protected sealed class FileLogger : IDisposable, ILogger
        {
            private StreamWriter mWritter;

            /// <summary>
            /// Creates a new instance of this class.
            /// </summary>
            /// <param name="path">The file path.</param>
            public FileLogger(string path)
            {
                try
                {
                    System.IO.File.Delete(path);
                }
                catch (Exception) { }

                try
                {
                    mWritter = System.IO.File.AppendText(path);
                }
                catch (Exception)
                {
                    // Bad, no loger available
                }
            }

            /// <summary>
            /// Object destructor.
            /// </summary>
            ~FileLogger()
            {
                Dispose();
            }

            /// <summary>
            /// Writes text line.
            /// </summary>
            /// <param name="s">the text</param>
            public void WriteLine(string s)
            {
                //Console.WriteLine(s);

                try
                {
                    mWritter.WriteLine(s);
                    mWritter.Flush();
                }
                catch (Exception)
                {
                    Dispose();
                }
            }

            /// <summary>
            /// Dispose object.
            /// </summary>
            public void Dispose()
            {
                if (mWritter != null)
                {
                    mWritter.Close();
                    mWritter = null;
                }
            }

        }

        // Helper class that implement input stream
        private class SerialPortInputStream : IInputStream
        {
            private SerialPort mSerialPort;

            public SerialPortInputStream(SerialPort serialPort)
            {
                this.mSerialPort = serialPort;
            }

            public int Read(byte[] buffer, int offset, int count)
            {
                int nbr = 0;

                if (mSerialPort.BytesToRead > 0)
                {
                    nbr = mSerialPort.Read(buffer, offset, count);
                    Datecs.Api.Util.Log.LogV("read(" + nbr + "): ", buffer, offset, nbr);
                }
                else
                {
                    //Datecs.Api.Util.Log.LogV("No data available");
                }

                return nbr;
            }
        }

        // Helper class that implements output stream 
        private class SerialPortOutputStream : IOutputStream
        {
            private SerialPort mSerialPort;

            public SerialPortOutputStream(SerialPort serialPort)
            {
                mSerialPort = serialPort;
            }

            public int Write(byte[] buffer, int offset, int count)
            {
                mSerialPort.Write(buffer, offset, count);
                Datecs.Api.Util.Log.LogV("write(" + count + "): ", buffer, offset, count);
                return count;
            }
        }

        // Hold instance of ProtocolAdapter object to handle printer in protocol mode.
        private ProtocolAdapter mProtocolAdapter;

        // Hold instance of Printer object to handle all printing tasks.
        private Printer mPrinter;

        // Hold instance of UniversalReader object to handle all atask relates with external periferials.
        private UniversalReader mUniversalReader;

        private FileLogger mLogger;

        public MainForm()
        {
            InitializeComponent();
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            UpdateSerialPort();
            UpdateBaudrate();
            
            // Set logger
            mLogger = new FileLogger(Application.ProductName + ".log");
            
            Log.SetLogger(mLogger);
            Log.LogD("--Start log--");
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            Properties.Settings.Default.SerialPort = cbSerialPort.Text;
            Properties.Settings.Default.BaudRate = cbBaudrate.Text;
            Properties.Settings.Default.Save();

            CloseActiveConnection();
            Log.LogD("--Stop log--");
            Log.SetLogger(null);
            mLogger.Dispose();
        }

        private void cbSerialPort_DropDown(object sender, EventArgs e)
        {
            UpdateSerialPort();
        }       
    
        // Helper method which remove garbage from Bluetooth port name
        private static string FixBlueToothPortNameBug(string portName)
        {
            if (!portName.StartsWith("COM"))
            {
                return portName;
            }

            string newPortName = "COM";                                // Start over with "COM" 
            foreach (char portChar in portName.Substring(3).ToCharArray())  //  Remove "COM", put the rest in a character array 
            {
                if (char.IsDigit(portChar))
                {
                    newPortName += portChar.ToString(); // Good character, append to portName                 
                }
            }

            return newPortName;
        }

        // Helper method to convert byte array to hex string
        private static string ByteArrayToHexString(byte[] data, int offset, int length) {
            char[] hex = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };
            char[] buf = new char[length * 2];
            int offs = 0;

            for (int i = 0; i < length; i++) {
                buf[offs++] = hex[(data[offset + i] >> 4) & 0xf];
                buf[offs++] = hex[(data[offset + i] >> 0) & 0xf];
            }

            return new string(buf, 0, offs);
        }

        // Helper method to convert byte array to hex string
        private static string ByteArrayToHexString(byte[] data) {
            return ByteArrayToHexString(data, 0, data.Length);
        }

        private void UpdateSerialPort()
        {
            string[] ports = System.IO.Ports.SerialPort.GetPortNames();
            for (int i = 0; i < ports.Length; i++)
            {
                ports[i] = FixBlueToothPortNameBug(ports[i]);
            }

            string port = Properties.Settings.Default.SerialPort;
            Array.Sort(ports, delegate(string s1, string s2)
            {
                int x = 0;
                int y = 0;
                int.TryParse(s1.Substring(3), out x);
                int.TryParse(s2.Substring(3), out y);
                return (x < y) ? -1 : ((x == y) ? 0 : 1);
            });

            cbSerialPort.Items.Clear();
            cbSerialPort.Items.AddRange(ports);
            cbSerialPort.SelectedIndex = cbSerialPort.Items.IndexOf(port);
        }

        private void UpdateBaudrate()
        {
            string baudrate = Properties.Settings.Default.BaudRate;

            cbBaudrate.SelectedIndex = cbBaudrate.Items.IndexOf(baudrate);
        }

        private void EstablishConnection()
        {
            serialPort.PortName = cbSerialPort.Text;
            serialPort.BaudRate = int.Parse(cbBaudrate.Text);

            Log.LogD("Open port " + serialPort.PortName + ", " + serialPort.BaudRate);
            serialPort.Open();

            IInputStream inputStream = new SerialPortInputStream(serialPort);
            IOutputStream outputStream = new SerialPortOutputStream(serialPort);
            try
            {
                Log.LogD("Get printer instance");

                // Check if printer is into protocol mode.
                mProtocolAdapter = new ProtocolAdapter(inputStream, outputStream);

                if (mProtocolAdapter.IsProtocolEnabled)
                {
                    mProtocolAdapter.OnPaperStateChanged += new ProtocolAdapter.PaperStateChangedEventHandler(ProtocolAdapter_OnPaperStateChanged);
                    mProtocolAdapter.OnReadCard += new ProtocolAdapter.ReadCardEventHandler(ProtocolAdapter_OnReadCard);

                    ProtocolAdapter.Channel printerChannel = mProtocolAdapter.GetChannel(ProtocolAdapter.ChannelType.Printer);
                    mPrinter = new Printer(printerChannel.GetInputStream(), printerChannel.GetOutputStream());                    

                    ProtocolAdapter.Channel ureaderChannel = mProtocolAdapter.GetChannel(ProtocolAdapter.ChannelType.UniversalReader);
                    mUniversalReader = new UniversalReader(ureaderChannel.GetInputStream(), ureaderChannel.GetOutputStream());
                }
                else
                {                    
                    // Device is not is protocol mode.
                    mPrinter = new Printer(mProtocolAdapter.GetRawInputStream(), mProtocolAdapter.GetRawOutputStream());

                    // In non protocol mode we can not access external devices.
                    mUniversalReader = null;
                }

                PrinterInformation pi = mPrinter.GetInformation();
                MessageBox.Show("Connected to " + pi.Name, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1);                

                mPrinter.Disconnect += (o, e) =>
                {
                    MessageBox.Show("Connection is closed");
                };
            }
            catch (Exception e)
            {
                Log.LogE("SDK Error: " + e.Message);
                CloseActiveConnection();
                throw e;
            }            
        }

        private void ProtocolAdapter_OnReadCard(bool encrypted)
        {
            this.Invoke((MethodInvoker)delegate()
            {
                if (encrypted)
                {
                    MessageBox.Show("Encrypted card data available");
                }
                else
                {
                    try
                    {
                        string[] tracks = mPrinter.ReadCard(true, true, true, 1000);

                        StringBuilder message = new StringBuilder();
                        message.Append("Card data: " + Environment.NewLine);
                        message.Append("Track 1: " + tracks[0] + Environment.NewLine);
                        message.Append("Track 2: " + tracks[1] + Environment.NewLine);
                        message.Append("Track 3: " + tracks[2] + Environment.NewLine);

                        MessageBox.Show(message.ToString(), this.Text, MessageBoxButtons.OK, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1);                
                    }
                    catch(IOException)
                    {
                        MessageBox.Show("Fail to read card", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);                
                    }
                }
            });
        }

        private void ProtocolAdapter_OnPaperStateChanged(bool hasPaper)
        {
            this.Invoke((MethodInvoker)delegate()
            {
                if (hasPaper)
                {
                    lblNoPaper.Visible = false;
                }
                else
                {
                    lblNoPaper.Visible = true;
                }
            });
        }

        private void CloseActiveConnection()
        {            
            if (mPrinter != null)
            {
                mPrinter.Dispose();
            }

            if (mUniversalReader != null)
            {
                mUniversalReader.Dispose();
            }

            if (mProtocolAdapter != null)
            {
                mProtocolAdapter.Dispose();
                mProtocolAdapter = null;
            }

            if (serialPort.IsOpen)
            {
                try
                {
                    serialPort.Close();
                }
                catch (Exception) {  }                
            }
        }

        private void ShowError(String text)
        {
            Log.LogE(text);
            MessageBox.Show(text, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button1);
        }

        private void btnPort_Click(object sender, EventArgs e)
        {            
            if (serialPort.IsOpen)
            {
                try
                {
                    CloseActiveConnection();
                    // Update UI
                    cbSerialPort.Enabled = true;
                    cbBaudrate.Enabled = true;
                    btnPort.Text = "Open";
                    panelContent.Enabled = false;
                }
                catch (Exception ex)
                {
                    ShowError("Failed to close: " + ex.Message);
                }
            }
            else
            {
                try
                {
                    EstablishConnection();
                    // Update UI
                    cbSerialPort.Enabled = false;
                    cbBaudrate.Enabled = false;
                    btnPort.Text = "Close";
                    panelContent.Enabled = true;
                }
                catch (Exception ex) 
                {
                    ShowError("Failed to open: " + ex.Message);
                }
            }
        }

        private void btnPrintSelfTest_Click(object sender, EventArgs e)
        {
            try
            {
                mPrinter.PrintSelfTest();
            }
            catch (System.IO.IOException ex)
            {
                ShowError("Failed to print self test: " + ex.Message);
            }
        }

        private void btnPrintText_Click(object sender, EventArgs e)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("{reset}{center}{w}{h}RECEIPT");
            sb.Append("{br}");
            sb.Append("{br}");
            sb.Append("{reset}1. {b}First item{br}");
            sb.Append("{reset}{right}{h}$0.50 A{br}");
            sb.Append("{reset}2. {u}Second item{br}");
            sb.Append("{reset}{right}{h}$1.00 B{br}");
            sb.Append("{reset}3. {i}Third item{br}");
            sb.Append("{reset}{right}{h}$1.50 C{br}");
            sb.Append("{br}");
            sb.Append("{reset}{right}{w}{h}TOTAL: {/w}$3.00  {br}");
            sb.Append("{br}");
            sb.Append("{reset}{center}{s}Thank You!{br}");

            try
            {                            
                mPrinter.Reset();
                mPrinter.PrintTaggedText(sb.ToString());
                mPrinter.FeedPaper(110);
            }
            catch (System.IO.IOException ex)
            {
                ShowError("Failed to print test: " + ex.Message);
            }
        }

        private void btnPrintImage_Click(object sender, EventArgs e)
        {
            Bitmap bitmap = Properties.Resources.SampleImage;
            int[] argb;

            // Specify bitmap format.
            PixelFormat pxf = PixelFormat.Format32bppRgb;

            // Lock the bitmap's bits.
            Rectangle rect = new Rectangle(0, 0, bitmap.Width, bitmap.Height);
            BitmapData bitmapData = bitmap.LockBits(rect, ImageLockMode.ReadOnly, pxf);

            // Get the address of the first line.
            IntPtr ptr = bitmapData.Scan0;

            // Copy the RGB values into the array.
            argb = new int[bitmap.Width * bitmap.Height];
            System.Runtime.InteropServices.Marshal.Copy(ptr, argb, 0, argb.Length);

            // Unlock the bits.
            bitmap.UnlockBits(bitmapData);

            try
            {
                mPrinter.Reset();
                mPrinter.PrintImage(argb, bitmap.Width, bitmap.Height, Printer.ALIGN_CENTER, true);
                mPrinter.FeedPaper(110);
            }
            catch (System.IO.IOException ex)
            {
                ShowError("Failed to print image: " + ex.Message);
            }
        }

        private void btnPrintPage_Click(object sender, EventArgs e)
        {
            try
            {
                PrinterInformation pi = mPrinter.GetInformation();

                if (pi.IsPageModeSupported == false)
                {
                    MessageBox.Show("Page mode is not supported by this device");
                    return;
                }

                mPrinter.Reset();
                mPrinter.SelectPageMode();

                mPrinter.SetPageRegion(0, 0, 160, 320, Printer.PAGE_LEFT);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH I{br}");
                mPrinter.DrawPageRectangle(0, 0, 160, 32, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from left to right" +
                        ", feed to bottom. Starting point in left top corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(160, 0, 160, 320, Printer.PAGE_TOP);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH II{br}");
                mPrinter.DrawPageRectangle(160 - 32, 0, 32, 320, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from top to bottom" +
                        ", feed to left. Starting point in right top corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(160, 320, 160, 320, Printer.PAGE_RIGHT);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH III{br}");
                mPrinter.DrawPageRectangle(0, 320 - 32, 160, 32, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from right to left" +
                        ", feed to top. Starting point in right bottom corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(0, 320, 160, 320, Printer.PAGE_BOTTOM);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH IV{br}");
                mPrinter.DrawPageRectangle(0, 0, 32, 320, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from bottom to top" +
                        ", feed to right. Starting point in left bottom corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(0, 0, 160, 320, Printer.PAGE_LEFT);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH I{br}");
                mPrinter.DrawPageRectangle(0, 0, 160, 32, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from left to right" +
                        ", feed to bottom. Starting point in left top corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(160, 0, 160, 320, Printer.PAGE_TOP);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH II{br}");
                mPrinter.DrawPageRectangle(160 - 32, 0, 32, 320, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from top to bottom" +
                        ", feed to left. Starting point in right top corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(160, 320, 160, 320, Printer.PAGE_RIGHT);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH III{br}");
                mPrinter.DrawPageRectangle(0, 320 - 32, 160, 32, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from right to left" +
                        ", feed to top. Starting point in right bottom corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(0, 320, 160, 320, Printer.PAGE_BOTTOM);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH IV{br}");
                mPrinter.DrawPageRectangle(0, 0, 32, 320, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from bottom to top" +
                        ", feed to right. Starting point in left bottom corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);
                mPrinter.SetPageRegion(0, 0, 160, 320, Printer.PAGE_LEFT);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH I{br}");
                mPrinter.DrawPageRectangle(0, 0, 160, 32, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from left to right" +
                        ", feed to bottom. Starting point in left top corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(160, 0, 160, 320, Printer.PAGE_TOP);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH II{br}");
                mPrinter.DrawPageRectangle(160 - 32, 0, 32, 320, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from top to bottom" +
                        ", feed to left. Starting point in right top corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(160, 320, 160, 320, Printer.PAGE_RIGHT);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH III{br}");
                mPrinter.DrawPageRectangle(0, 320 - 32, 160, 32, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from right to left" +
                        ", feed to top. Starting point in right bottom corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(0, 320, 160, 320, Printer.PAGE_BOTTOM);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH IV{br}");
                mPrinter.DrawPageRectangle(0, 0, 32, 320, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from bottom to top" +
                        ", feed to right. Starting point in left bottom corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);
                mPrinter.SetPageRegion(0, 0, 160, 320, Printer.PAGE_LEFT);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH I{br}");
                mPrinter.DrawPageRectangle(0, 0, 160, 32, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from left to right" +
                        ", feed to bottom. Starting point in left top corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(160, 0, 160, 320, Printer.PAGE_TOP);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH II{br}");
                mPrinter.DrawPageRectangle(160 - 32, 0, 32, 320, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from top to bottom" +
                        ", feed to left. Starting point in right top corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(160, 320, 160, 320, Printer.PAGE_RIGHT);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH III{br}");
                mPrinter.DrawPageRectangle(0, 320 - 32, 160, 32, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from right to left" +
                        ", feed to top. Starting point in right bottom corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(0, 320, 160, 320, Printer.PAGE_BOTTOM);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH IV{br}");
                mPrinter.DrawPageRectangle(0, 0, 32, 320, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from bottom to top" +
                        ", feed to right. Starting point in left bottom corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);
                mPrinter.SetPageRegion(0, 0, 160, 320, Printer.PAGE_LEFT);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH I{br}");
                mPrinter.DrawPageRectangle(0, 0, 160, 32, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from left to right" +
                        ", feed to bottom. Starting point in left top corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(160, 0, 160, 320, Printer.PAGE_TOP);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH II{br}");
                mPrinter.DrawPageRectangle(160 - 32, 0, 32, 320, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from top to bottom" +
                        ", feed to left. Starting point in right top corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(160, 320, 160, 320, Printer.PAGE_RIGHT);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH III{br}");
                mPrinter.DrawPageRectangle(0, 320 - 32, 160, 32, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from right to left" +
                        ", feed to top. Starting point in right bottom corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(0, 320, 160, 320, Printer.PAGE_BOTTOM);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH IV{br}");
                mPrinter.DrawPageRectangle(0, 0, 32, 320, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from bottom to top" +
                        ", feed to right. Starting point in left bottom corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);
                mPrinter.SetPageRegion(0, 0, 160, 320, Printer.PAGE_LEFT);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH I{br}");
                mPrinter.DrawPageRectangle(0, 0, 160, 32, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from left to right" +
                        ", feed to bottom. Starting point in left top corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(160, 0, 160, 320, Printer.PAGE_TOP);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH II{br}");
                mPrinter.DrawPageRectangle(160 - 32, 0, 32, 320, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from top to bottom" +
                        ", feed to left. Starting point in right top corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(160, 320, 160, 320, Printer.PAGE_RIGHT);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH III{br}");
                mPrinter.DrawPageRectangle(0, 320 - 32, 160, 32, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from right to left" +
                        ", feed to top. Starting point in right bottom corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);

                mPrinter.SetPageRegion(0, 320, 160, 320, Printer.PAGE_BOTTOM);
                mPrinter.SetPageXY(0, 4);
                mPrinter.PrintTaggedText("{reset}{center}{b}PARAGRAPH IV{br}");
                mPrinter.DrawPageRectangle(0, 0, 32, 320, Printer.FILL_INVERTED);
                mPrinter.SetPageXY(0, 34);
                mPrinter.PrintTaggedText("{reset}Text printed from bottom to top" +
                        ", feed to right. Starting point in left bottom corner of the page.{br}");
                mPrinter.DrawPageFrame(0, 0, 160, 320, Printer.FILL_BLACK, 1);
                mPrinter.PrintPage();
                mPrinter.SelectStandardMode();
                mPrinter.FeedPaper(110);
            }
            catch (System.IO.IOException ex)
            {
                ShowError("Failed to print page: " + ex.Message);
            } 
        }

        private void btnPrintBarcode_Click(object sender, EventArgs e)
        {
            try
            {
                mPrinter.Reset();

                mPrinter.SetBarcode(Printer.ALIGN_CENTER, false, 2, Printer.HRI_BELOW, 100);
                mPrinter.PrintBarcode(Printer.BARCODE_EAN13, "123456789012");
                mPrinter.FeedPaper(38);

                mPrinter.SetBarcode(Printer.ALIGN_CENTER, false, 2, Printer.HRI_BOTH, 100);
                mPrinter.PrintBarcode(Printer.BARCODE_CODE128, "ABCDEF123456");
                mPrinter.FeedPaper(38);

                mPrinter.SetBarcode(Printer.ALIGN_CENTER, false, 2, Printer.HRI_NONE, 100);
                mPrinter.PrintBarcode(Printer.BARCODE_PDF417, "ABCDEFGHIJKLMNOPQRSTUVWXYZ");
                mPrinter.FeedPaper(38);

                mPrinter.SetBarcode(Printer.ALIGN_CENTER, false, 2, Printer.HRI_NONE, 100);
                mPrinter.PrintQRCode(4, 3, "http://www.datecs.bg");
                mPrinter.FeedPaper(38);

                mPrinter.FeedPaper(110);
            }
            catch (System.IO.IOException ex)
            {
                ShowError("Failed to print barcode: " + ex.Message);
            }
        }

        private void btnSetClock_Click(object sender, EventArgs e)
        {
            try
            {
                mPrinter.SetRTC(DateTime.Now);                
            }
            catch (System.IO.IOException ex)
            {
                ShowError("Failed set clock: " + ex.Message);
            }
        }

        private void btnReadMsCard_Click(object sender, EventArgs e)
        {
            String[] tracks = null;
            FinancialCard card = null;

            try
            {
                PrinterInformation pi = mPrinter.GetInformation();

                if (pi.Name.StartsWith("CMP-10"))
                {
                    // The printer CMP-10 can read only two tracks at once.
                    tracks = mPrinter.ReadCard(true, true, false, 15000);
                }
                else
                {
                    tracks = mPrinter.ReadCard(true, true, true, 15000);
                }
            }
            catch (System.IO.IOException ex)
            {
                ShowError("Failed to read card: " + ex.Message);
            }

            if (tracks != null)
            {
                StringBuilder msg = new StringBuilder();

                if (tracks[0] == null && tracks[1] == null && tracks[2] == null)
                {
                    msg.Append("No card read");
                }
                else
                {
                    if (tracks[0] != null)
                    {
                        card = new FinancialCard(tracks[0]);
                    }
                    else if (tracks[1] != null)
                    {
                        card = new FinancialCard(tracks[1]);
                    }

                    if (card != null)
                    {
                        msg.Append("Number: " + card.GetNumber());
                        msg.Append("\n");
                        msg.Append("Holder: " + card.GetName());
                        msg.Append("\n");
                        msg.Append("Expiry date: " + card.GetExpiryMonth().ToString("D2") + "/" + card.GetExpiryYear().ToString("D2"));
                        msg.Append("\n");
                    }

                    if (tracks[0] != null)
                    {
                        msg.Append("\n");
                        msg.Append(tracks[0]);

                    }
                    if (tracks[1] != null)
                    {
                        msg.Append("\n");
                        msg.Append(tracks[1]);
                    }
                    if (tracks[2] != null)
                    {
                        msg.Append("\n");
                        msg.Append(tracks[2]);
                    }
                }

                MessageBox.Show(msg.ToString(), "Card info", MessageBoxButtons.OK, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1);
            }
        }

        private void btnReadSmartCard_Click(object sender, EventArgs e)
        {
            if (mUniversalReader == null)
            {                
                MessageBox.Show("Printer is not in protocol mode", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button1);
                return;
            }

            try
            {
                UniversalReader.SmartCardReader scr = mUniversalReader.GetSmartCardReader();
                bool cardAvailable = scr.IsCardPresent();

                if (cardAvailable)
                {
                    // Get Answer-To-Reset string
                    AnswerToReset atr = scr.Reset();

                    StringBuilder msg = new StringBuilder();
                    msg.Append("ATR: " + ByteArrayToHexString(atr.GetData()) + "\n");

                    // Try to perform GET CHALLANGE  APDU command
                    //ResponseAPDU response = scr.Transmit(0x00, 0x84, 0, 0, null);
                    //if (response.GetSW() == 0x9000)
                    //{
                    //    msg.Append("RND: " + ByteArrayToHexString(response.GetData()) + "\n");
                    //}
                    
                    MessageBox.Show(msg.ToString(), "Smart card info", MessageBoxButtons.OK, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button1);

                }
                else
                {
                    MessageBox.Show("No smart card available", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1);
                }        
            }
            catch (System.IO.IOException ex)
            {
                ShowError("Failed to read smart card: " + ex.Message);
            }
        }        

    }
}



      