﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.IO.Ports;
using System.Security.Cryptography;
using System.Text;
using System.Windows.Forms;

using Datecs.Api.IO;
using Datecs.Api.Card;
using Datecs.Api.Linea;
using Datecs.Api.Printer;
using Datecs.Api.UniversalReader;
using Datecs.Api.SmartCard;
using Datecs.Api.Util;

namespace LineaProDemo
{
    public partial class MainForm : Form
    {
        private readonly byte[] DUKPT_BDK = new byte[] { 0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF, 0xFE, 0xDC, 0xBA, 0x98, 0x76, 0x54, 0x32, 0x10 };       

        // Helper class that implement input stream
        private class SerialPortInputStream : IInputStream
        {
            private SerialPort mSerialPort;

            public SerialPortInputStream(SerialPort serialPort)
            {
                this.mSerialPort = serialPort;
            }

            public int Read(byte[] buffer, int offset, int count)
            {
                int nbr = 0;

                if (mSerialPort.BytesToRead > 0)
                {
                    nbr = mSerialPort.Read(buffer, offset, count);

                    //Debug("Read(" + nbr + "): " + ByteArrayToHexString(buffer, offset, nbr));                    
                }
                else
                {
                    // System.Diagnostics.Debug.WriteLine("Read(0)");
                }
                Debug("Read(" + nbr + "): " + ByteArrayToHexString(buffer, offset, nbr));            
                return nbr;
            }
        }

        // Helper class that implements output stream 
        private class SerialPortOutputStream : IOutputStream
        {
            private SerialPort mSerialPort;

            public SerialPortOutputStream(SerialPort serialPort)
            {
                mSerialPort = serialPort;
            }

            public int Write(byte[] buffer, int offset, int count)
            {
                try
                {
                    mSerialPort.Write(buffer, offset, count);
                    Debug("Write(" + count + "): " + ByteArrayToHexString(buffer, offset, count));
                    return count;
                }
                catch (IOException) 
                {
                    return 0;
                }                
            }
        }

        private LineaPro mLineaPro = null;

        private bool mIsActive = false;

        public MainForm()
        {
            InitializeComponent();
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            UpdateSerialPort();
            UpdateBaudrate();
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            CloseConnection();
            Properties.Settings.Default.SerialPort = cbSerialPort.Text;
            Properties.Settings.Default.BaudRate = cbBaudrate.Text;
            Properties.Settings.Default.Save();
        }

        private void cbSerialPort_DropDown(object sender, EventArgs e)
        {
            UpdateSerialPort();
        }

        // Helper method for debuging.
        private static void Debug(string text)
        {
            string dateString = DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss.fff");

            Console.WriteLine(dateString + " " + text);
        }

        // Helper method which remove garbage from Bluetooth port name
        private static string FixBlueToothPortNameBug(string portName)
        {
            if (!portName.StartsWith("COM"))
            {
                return portName;
            }

            string newPortName = "COM";                                // Start over with "COM" 
            foreach (char portChar in portName.Substring(3).ToCharArray())  //  Remove "COM", put the rest in a character array 
            {
                if (char.IsDigit(portChar))
                {
                    newPortName += portChar.ToString(); // Good character, append to portName                 
                }
            }

            return newPortName;
        }

        // Helper method to convert byte array to hex string
        private static string ByteArrayToHexString(byte[] data, int offset, int length)
        {
            char[] hex = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };
            char[] buf = new char[length * 2];
            int offs = 0;

            for (int i = 0; i < length; i++)
            {
                buf[offs++] = hex[(data[offset + i] >> 4) & 0xf];
                buf[offs++] = hex[(data[offset + i] >> 0) & 0xf];
            }

            return new string(buf, 0, offs);
        }

        // Helper method to convert byte array to hex string
        private static string ByteArrayToHexString(byte[] data)
        {
            return ByteArrayToHexString(data, 0, data.Length);
        }

        private void UpdateSerialPort()
        {
            string[] ports = System.IO.Ports.SerialPort.GetPortNames();
            for (int i = 0; i < ports.Length; i++)
            {
                ports[i] = FixBlueToothPortNameBug(ports[i]);
            }

            string port = Properties.Settings.Default.SerialPort;
            Array.Sort(ports, delegate(string s1, string s2)
            {
                int x = 0;
                int y = 0;
                int.TryParse(s1.Substring(3), out x);
                int.TryParse(s2.Substring(3), out y);
                return (x < y) ? -1 : ((x == y) ? 0 : 1);
            });

            cbSerialPort.Items.Clear();
            cbSerialPort.Items.AddRange(ports);
            cbSerialPort.SelectedIndex = cbSerialPort.Items.IndexOf(port);
        }

        private void UpdateBaudrate()
        {
            string baudrate = Properties.Settings.Default.BaudRate;

            cbBaudrate.SelectedIndex = cbBaudrate.Items.IndexOf(baudrate);
        }

        private void EstablishConnection()
        {
            txtLog.Text = string.Empty;

            // Open serial port
            try
            {
                serialPort.PortName = cbSerialPort.Text;
                serialPort.BaudRate = int.Parse(cbBaudrate.Text);
                serialPort.Open();
            }
            catch (Exception ex)
            {
                MessageBox.Show("Failed to open serial port: " + ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
                return;
            }
            
            IInputStream inputStream = new SerialPortInputStream(serialPort);
            IOutputStream outputStream = new SerialPortOutputStream(serialPort);
            try
            {
                Debug("Get Linea Pro instance");                
                mLineaPro = new LineaPro(inputStream, outputStream);
                this.mLineaPro.Beep(100, 2400, 100);

                // Get device information
                LineaPro.Information info = this.mLineaPro.GetInformation();
                AddText("Name: " + info.Name);
                AddText("Model: " + info.Model);
                AddText("Version: " + info.Version);
                AddText("Manifacturer: " + info.Manifaturer);
                AddText("Serial number: " + info.SerialNumber);

                // Get battery information
                int batteryLevel = this.mLineaPro.GetBatteryLevel();
                AddText("Battery level: " + batteryLevel + "%");

                // Enable barcode scanner
                this.mLineaPro.bcSetMode(LineaPro.BARCODE_MODE_MULTI_SCAN);
                this.mLineaPro.bcEnableScanButton(true);

                if (info.HasMagneticHead)
                {
                    // Enable and set encrypted magnetic head                    
                    if (info.HasEncryptedMagneticHead)
                    {
                        this.mLineaPro.emsrEnable(true);
                        this.mLineaPro.emsrSetEncryptionAlgorithm(LineaPro.ENCRYPTION_ALGORITHM_EH_IDTECH);
                    }
                    else
                    {
                        this.mLineaPro.msrEnable(true);
                    }
                }

                // Set listeners 
                this.mLineaPro.ButtonPressed += LineaPro_ButtonPressed;
                this.mLineaPro.BarcodeReceived += LineaPro_BarcodeReceived;
                this.mLineaPro.MagneticCardDataReceived += LineaPro_MagneticCardDataReceived;
                this.mLineaPro.Disconnected += mLineaPro_Disconnected;

                AddText("-------------------------------");

                LineaPro.Information lpi = mLineaPro.GetInformation();
                MessageBox.Show("Connected to " + lpi.Name, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1);
            }
            catch (IOException ex)
            {
                MessageBox.Show("Failed to connect to device: " + ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
                CloseConnection();
                return;
            }

            mIsActive = true;
            cbSerialPort.Enabled = false;
            cbBaudrate.Enabled = false;
            btnPort.Text = "Close";
            txtLog.Enabled = true;            
        }

        private void CloseConnection()
        {            
            if (mLineaPro != null)
            {
                mLineaPro.Dispose();
            }

            if (serialPort.IsOpen)
            {
                serialPort.Close();
            }

            mIsActive = false;
            cbSerialPort.Enabled = true;
            cbBaudrate.Enabled = true;
            btnPort.Text = "Open";
            txtLog.Enabled = false;
        }
        
        private void LineaPro_ButtonPressed(object sender, LineaPro.ButtonEventArgs e)
        {
            // Run on UI Thread
            this.BeginInvoke(new MethodInvoker(delegate()
            {
                if (e.BarcodeSwitch)
                {
                    Debug("Barcode switch is pressed");
                    txtLog.BackColor = Color.FromArgb(255, 0, 0);
                }
                else
                {
                    Debug("Barcode switch is released");
                    txtLog.BackColor = SystemColors.Window;
                }
            }));
        }

        private void LineaPro_BarcodeReceived(object sender, LineaPro.BarcodeEventArgs e)
        {
            Debug("Barcode is received (Type=" + e.Type + ",Barcode=" + e.DataString + ")");

            AddText("Barcode: " + e.DataString);
            AddText("-------------------------------");
        }

        private void LineaPro_MagneticCardDataReceived(object sender, LineaPro.MagneticCardDataEventArgs e)
        {
            // Play beep of card read
            try
            {
                this.mLineaPro.Beep(100, 2400, 200);
            }
            catch (IOException) { }

            StringBuilder sb = new StringBuilder();

            foreach (char c in System.Text.Encoding.UTF8.GetChars(e.Data))
            {
                if (c >= 32 && c <= 127)
                {
                    sb.Append(c);
                }
                else
                {
                    sb.Append('.');
                }
            }

            Debug("Magnetic card data is received (IsEncrypted=" + e.IsEncrypted + ", IsRaw=" + e.IsRaw + ", Data=" + sb.ToString() + ")");

            if (e.IsEncrypted)
            {
                if (e.EncryptionAlgorithm == LineaPro.ENCRYPTION_ALGORITHM_EH_IDTECH)
                {
                    DecodeIDTechBlock(DUKPT_BDK, e.Data);
                }
                else
                {
                    AddText("Card data: " + ByteArrayToHexString(e.Data, 0, e.Data.Length));
                }
            }
            else if (e.IsPlainText)
            {
                AddText("Card data: " + System.Text.Encoding.UTF8.GetString(e.Data, 1, e.Data.Length - 1));
            }

            AddText("-------------------------------");
        }

        private void mLineaPro_Disconnected(object sender, EventArgs e)
        {
            // Run on UI Thread
            this.BeginInvoke(new MethodInvoker(delegate()
            {                
                CloseConnection();                
            }));
        }        

        private void DecodeIDTechBlock(byte[] bdk, byte[] data)
        {
            /* IDTech format:
             * DATA[0]:	CARD TYPE: 0x0 - payment card
             * DATA[1]:	TRACK FLAGS
             * DATA[2]:	TRACK 1 LENGTH
             * DATA[3]:	TRACK 2 LENGTH
             * DATA[4]:	TRACK 3 LENGTH
             * DATA[??]: TRACK 1 DATA MASKED
             * DATA[??]: TRACK 2 DATA MASKED
             * DATA[??]: TRACK 3 DATA
             * DATA[??]: TRACK 1 AND TRACK 2 TDES ENCRYPTED
             * DATA[??]: TRACK 1 SHA1 (0x14 BYTES)
             * DATA[??]: TRACK 2 SHA1 (0x14 BYTES)
             * DATA[??]: DUKPT SERIAL AND COUNTER (0x0A BYTES)
             */
            int cardType = data[0] & 0xff;
            int track1Len = data[2] & 0xff;
            int track2Len = data[3] & 0xff;
            int track3Len = data[4] & 0xff;
            int offset = 5;

            AddText("Card type: " + ((cardType == 0) ? "PAYMENT" : "UNKNOWN"));

            if (track1Len > 0)
            {
                AddText("Track1(M): " + System.Text.Encoding.UTF8.GetString(data, offset, track1Len));
                offset += track1Len;
            }

            if (track2Len > 0)
            {
                AddText("Track2(M): " + System.Text.Encoding.UTF8.GetString(data, offset, track2Len));
                offset += track2Len;
            }

            if (track3Len > 0)
            {
                AddText("Track3: " + System.Text.Encoding.UTF8.GetString(data, offset, track3Len));
                offset += track3Len;
            }

            if ((track1Len + track2Len) > 0)
            {
                int blockSize = (track1Len + track2Len + 7) & 0xFFFFFF8;
                byte[] encrypted = new byte[blockSize];
                Array.Copy(data, offset, encrypted, 0, encrypted.Length);
                offset += blockSize;

                byte[] track1Hash = null;
                if (track1Len > 0)
                {
                    track1Hash = new byte[20];
                    Array.Copy(data, offset, track1Hash, 0, track1Hash.Length);
                }
                offset += 20;

                byte[] track2Hash = null;
                if (track2Len > 0)
                {
                    track2Hash = new byte[20];
                    Array.Copy(data, offset, track2Hash, 0, track2Hash.Length);
                }
                offset += 20;

                byte[] ksn = new byte[10];
                Array.Copy(data, offset, ksn, 0, ksn.Length);

                byte[] ipek = Crypto.CalculateIPEK(bdk, ksn);
                byte[] dataKey = Crypto.CalculateDataKey(ksn, ipek);
                byte[] decrypted = Crypto.Decrypt3DESCBC(dataKey, encrypted);

                Debug("ksn: " + ByteArrayToHexString(ksn, 0, ksn.Length));
                Debug("ipek: " + ByteArrayToHexString(ipek, 0, ipek.Length));
                Debug("dataKey: " + ByteArrayToHexString(dataKey, 0, dataKey.Length));
                Debug("encrypted: " + ByteArrayToHexString(encrypted, 0, encrypted.Length));
                Debug("decrypted: " + ByteArrayToHexString(decrypted, 0, decrypted.Length));

                if (track1Hash != null)
                {
                    SHA1Managed digest = new SHA1Managed();
                    digest.ComputeHash(decrypted, 0, track1Len);
                    if (!ArrayEqual(track1Hash, digest.Hash))
                    {
                        AddText("Failed to decrypt track 1");
                        track1Len = 0;
                    }
                }

                if (track2Hash != null)
                {
                    SHA1Managed digest = new SHA1Managed();
                    digest.ComputeHash(decrypted, track1Len, track2Len);
                    if (!ArrayEqual(track2Hash, digest.Hash))
                    {
                        AddText("Failed to decrypt track 2");
                        track2Len = 0;
                    }
                }

                if (track1Len > 0)
                {
                    AddText("Track1: " + System.Text.Encoding.UTF8.GetString(decrypted, 0, track1Len));
                }

                if (track2Len > 0)
                {
                    AddText("Track2: " + System.Text.Encoding.UTF8.GetString(decrypted, track1Len, track2Len));
                }
            }
        }
        
        private bool ArrayEqual(byte[] a, byte[] b)
        {
            if (a != b)
            {
                if (a.Length != b.Length) return false;

                for (int i = 0; i < a.Length; i++)
                {
                    if (a[i] != b[i]) return false;
                }
            }

            return true;
        }

        private void AddText(string text)
        {
            if (txtLog.InvokeRequired)
            {
                txtLog.BeginInvoke(new MethodInvoker(delegate()
                {
                    txtLog.AppendText(text + Environment.NewLine);
                }));
            }
            else
            {
                txtLog.AppendText(text + Environment.NewLine);
            }
        }

        private void btnPort_Click(object sender, EventArgs e)
        {
            if (mIsActive)
            {
                CloseConnection();
            }
            else
            {
                EstablishConnection();
            }
        }
    
    }
}
