﻿using System;
using System.Drawing;
using System.Drawing.Imaging;

namespace PinpadDemo
{
    class ImageHelper
    {        
        /// <summary>
        /// Converts RGB color array to gray color array.
        /// </summary>
        /// <param name="argb">The RGB data buffer.</param>
        /// <param name="width">The image width.</param>
        /// <param name="height">The image height.</param>
        public static void ConvertARGBToGrayscale(int[] argb, int width, int height) 
        {
            int pixels = width * height;
        
            for (int i = 0; i < pixels; i++) 
            {
                int r = (argb[i] >> 16) & 0xff;
                int g = (argb[i] >> 8) & 0xff;
                int b = argb[i] & 0xff;
                int color = ((r * 19 + g * 38 + b * 7) >> 6) & 0xff;
                argb[i] = color;            
            }
        }

        /// <summary>
        /// Perform error diffusion using Floyd-Stainberg algorithm.
        /// </summary>
        /// <param name="grayscale">The gray data buffer.</param>
        /// <param name="width">The image width.</param>
        /// <param name="height">The image height.</param>
        public static void DitherImageByFloydSteinberg(int[] grayscale, int width, int height) 
        {
            int v, error, ed;
            int stopXM1 = width - 1;
            int stopYM1 = height - 1;
            int[] coef = new int[] { 3, 5, 1 };
                    
            for (int y = 0, offs = 0; y < height; y++ ) 
            {
                for (int x = 0; x < width; x++, offs++) 
                {
                    v = grayscale[offs];

                    // fill the next destination pixel
                    if (v < 128) {
                        grayscale[offs] = 0;
                        error = v;
                    } 
                    else 
                    {
                        grayscale[offs] = 255;
                        error = v - 255;
                    }

                    // calculate error diffusion
                    if (x != stopXM1) 
                    {
                        // right pixel
                        ed = grayscale[offs + 1] + (error * 7) / 16;
                        ed = (ed < 0) ? 0 : (( ed > 255) ? 255 : ed);
                        grayscale[offs + 1] = ed;
                    }

                    if (y != stopYM1) 
                    {
                        // bottom pixels                    
                        for (int i = -1, j = 0; i <= 1; i++, j++) 
                        {
                            if ((x + i >= 0) && (x + i < width)) 
                            {
                                ed = grayscale[offs + width + i] + (error * coef[j]) / 16;
                                ed = (ed < 0) ? 0 : (( ed > 255) ? 255 : ed);
                                grayscale[offs + width + i] = ed;
                            }
                        }
                    }
                }            
            }
        }

        /// <summary>
        /// Get RGB array from bitmap.
        /// </summary>
        /// <param name="bitmap">The bitmap.</param>
        /// <returns></returns>
        public static int[] GetArgb(Bitmap bitmap)
        {
            int[] argb;

            // Specify bitmap format.
            PixelFormat pxf = PixelFormat.Format32bppRgb;

            // Lock the bitmap's bits.
            Rectangle rect = new Rectangle(0, 0, bitmap.Width, bitmap.Height);
            BitmapData bitmapData = bitmap.LockBits(rect, ImageLockMode.ReadOnly, pxf);

            // Get the address of the first line.
            IntPtr ptr = bitmapData.Scan0;

            // Copy the RGB values into the array.
            argb = new int[bitmap.Width * bitmap.Height];
            System.Runtime.InteropServices.Marshal.Copy(ptr, argb, 0, argb.Length);
            
            // Unlock the bits.
            bitmap.UnlockBits(bitmapData);

            return argb;
        }

        /// <summary>
        /// Get monochrome array from bitmap.
        /// </summary>
        /// <param name="bitmap">The bitmap.</param>
        /// <param name="dither">Perform error diffusion.</param>
        /// <returns></returns>
        public static int[] GetMonochromeData(Bitmap bitmap, bool dither)
        {
            int[] argb = GetArgb(bitmap);

            ConvertARGBToGrayscale(argb, bitmap.Width, bitmap.Height);

            if (dither)
            {
                DitherImageByFloydSteinberg(argb, bitmap.Width, bitmap.Height);
            }

            return argb;
        }
    }
}
